import json
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import typer

from invgen.hosts import GeneratedHost, get_all_generated_hosts
from invgen.logging import init_logger, logger


@dataclass
class Group:
    name: str
    hosts: list[str]


class AnsibleInventory:
    def __init__(self, hosts: list[GeneratedHost]):
        self.hosts = hosts

    def _build_groups(self) -> list[Group]:
        group_filter: dict[str, list[str]] = defaultdict(list)

        for host in self.hosts:
            if "metadata" in host.vars:
                for k, v in host.vars["metadata"].items():
                    if isinstance(v, list):
                        for item in v:
                            group_filter[f"{k}_{item}"].append(host.name)
                    elif isinstance(v, str):
                        group_filter[f"{k}_{v}"].append(host.name)
                    else:
                        logger.warning(
                            f"Invalid metadata type {type(v)} ({k}/{v}) on {host.name}"
                        )
            else:
                logger.warning(f"Host {host.name} has no metadata")
                group_filter["ungrouped"].append(host.name)

        groups = [Group(name=k, hosts=v) for k, v in group_filter.items()]
        return groups

    def _build_hostvars(self) -> dict[str, dict[str, Any]]:
        return {host.name: host.vars for host in self.hosts}

    def build(self) -> dict[str, Any]:
        """Builds the inventory"""
        groups = self._build_groups()
        inventory: dict[str, Any] = {}
        inventory["_meta"] = {"hostvars": self._build_hostvars()}
        all_hosts = set()

        for group in groups:
            inventory[group.name] = {"hosts": group.hosts}
            all_hosts.update(group.hosts)

        inventory["all"] = {"hosts": list(all_hosts)}
        return inventory

    def render(self, pretty: bool = False) -> str:
        """Renders the inventory as json"""
        if pretty:
            return json.dumps(self.build(), indent=2)

        return json.dumps(self.build())

    def build_host(self, host: str) -> dict[str, Any]:
        """Builds hostvars for a specific host"""
        hostvars = self._build_hostvars()
        return hostvars.get(host, {})

    def render_host(self, host: str, pretty: bool = False) -> str:
        if pretty:
            return json.dumps(self.build_host(host), indent=2)
        return json.dumps(self.build_host(host))

    def __str__(self):
        return str(self.hosts)


inventory_app = typer.Typer()


@inventory_app.command()
def inventory(
    source: Path = typer.Option(
        Path().cwd(), envvar="INVGEN_SOURCE", help="Source directory"
    ),
    pretty: bool = False,
    list_hosts: bool = typer.Option(False, "--list", help="Output inventory"),
    host: str = typer.Option("", "--host", help="Output hostvars for a host"),
    log_level: str = typer.Option("INFO", help="Log level"),
):
    init_logger(log_level)

    logger.info(f"Generating inventory from {source}")
    hosts = get_all_generated_hosts(source)
    inventory = AnsibleInventory(hosts)

    if list_hosts:
        print(inventory.render(pretty=pretty))
    elif len(host) > 0:
        print(inventory.render_host(host, pretty=pretty))
    else:
        typer.echo("No command specified (--host or --list)", err=True)
        raise SystemExit(1)


if __name__ == "__main__":
    inventory_app()



import typer
from pathlib import Path

from invgen.logging import init_logger
from invgen.hosts import generate_hosts
from invgen.inventory import inventory_app
from invgen.templates import render_template
from invgen.watcher import watch_for_changes

app = typer.Typer()
app_new = typer.Typer()
app.add_typer(inventory_app, name="inventory")
app.add_typer(app_new, name="new")


@app.command()
def generate(
    source: Path = typer.Option(
        Path().cwd(), "-s", "--source", envvar="INVGEN_SOURCE", help="Source directory"
    ),
    verbose: bool = False,
    debug: bool = False,
    watch: bool = typer.Option(False, "-w", "--watch", help="Watch for changes"),
):
    if verbose:
        init_logger("INFO")
    elif debug:
        init_logger("DEBUG")
    else:
        init_logger()

    typer.echo(f"=> Generating hosts from {source}/hosts/")
    generate_hosts(source)
    typer.echo(f"=> Done! Generated hosts in {source}/generated/")

    if watch:
        watch_for_changes(source)


@app_new.command(name="host")
def new_host(
    destination: Path = typer.Option(
        Path().cwd() / "hosts/",
        "-d",
        "--destination",
        envvar="INVGEN_DESTINATION",
        help="Destination directory",
    ),
    template: Path = typer.Option(
        Path("templates/host.yaml"),
        "-t",
        "--template",
        envvar="INVGEN_HOST_TEMPLATE",
        help="Path to host template",
    ),
    name: str = typer.Option(..., help="Name of the new host"),
    options: str = typer.Option(
        "",
        "-o",
        "--options",
        help="Additional options as key=value pairs, separated by spaces. "
        + "Can be a list seperated by commas.",
    ),
    ignore_errors: bool = typer.Option(
        False, "--ignore-errors", help="Ignore errors in the template"
    ),
):
    typer.echo(f"=> Creating new host {name} from {template}")

    kwargs: dict[str, str | list] = {"name": name}
    if options:
        for option in options.split(" "):
            key, value = option.split("=")
            if "," in value:
                kwargs[key.strip()] = value.split(",")
            else:
                kwargs[key.strip()] = value.strip()
    else:
        typer.echo(typer.style("=> No options provided.", fg=typer.colors.YELLOW))

    rendered = render_template(template, undefined_error=not ignore_errors, **kwargs)
    with open(destination / f"{name}.yaml", "w") as f:
        f.write(rendered)

    typer.echo(f"=> Done! Created new host in {destination}/{name}.yaml")



import logging

logger = logging.getLogger("invgen")


def init_logger(loglevel: str = "WARNING") -> None:
    default_format = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(funcName)s - %(message)s"
    )
    default_handler = logging.StreamHandler()
    default_handler.setLevel(loglevel)
    default_handler.setFormatter(default_format)

    logger.addHandler(default_handler)
    logger.setLevel("DEBUG")



from pathlib import Path
from time import sleep

import typer
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from invgen.hosts import generate_hosts


def watch_for_changes(source: Path):
    class RegenerateHandler(FileSystemEventHandler):
        def _run(self, event):
            if (
                "generated" not in event.src_path
                and event.src_path.endswith(".yaml")
                and event.is_directory is False
            ):
                print(f"=> File {event.event_type}: {event.src_path}")
                generate_hosts(source)
                print(f"=> Done! Regenerated hosts in {source}/generated/")

        def on_created(self, event):
            self._run(event)

        def on_modified(self, event):
            self._run(event)

        def on_deleted(self, event):
            self._run(event)

    event_handler = RegenerateHandler()
    observer = Observer()
    observer.schedule(event_handler, str(source.joinpath("hosts/")), recursive=True)
    observer.schedule(event_handler, str(source.joinpath("metadata/")), recursive=True)
    observer.start()

    typer.echo("=> Watching for changes... Press Ctrl+C to stop.")
    try:
        while True:
            sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryFile

import yaml

from invgen.files import load_yaml, load_yaml_cached, save_yaml
from invgen.logging import logger
from invgen.metadata import MetadataVars, build_metadata_vars


def generate_hosts(data_dir: Path):
    metadata = build_metadata_vars(data_dir)
    files = get_all_host_files(data_dir)
    logger.info(f"Generating {len(files)} hosts")
    for host in files:
        logger.info(f"Generating host {host.stem}")
        path = get_generated_host_path(data_dir, host.stem)
        content = generate_host_file(host, metadata)

        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            f.write(content)


@dataclass
class ValueWithSource:
    value: dict
    source: str


def generate_host_file(host: Path, metadata: MetadataVars) -> str:
    """Generate the content of a host file based on the provided metadata."""
    host_vars = load_yaml_cached(host)
    if "metadata" not in host_vars:
        logger.warning(f"Host {host.stem} has no metadata")
        host_vars["metadata"] = {}

    host_vars_struct: dict[str, ValueWithSource] = {}

    with TemporaryFile("w+") as f:
        for metadata_type, metadata_value in host_vars["metadata"].items():
            if isinstance(metadata_value, str):
                logger.debug(f"Processing metadata {metadata_type}/{metadata_value}")
                source = f"{metadata_type}/{metadata_value}"
                if metadata.lookup(metadata_type, host_vars["metadata"][metadata_type]):
                    for k, v in metadata.lookup(
                        metadata_type, host_vars["metadata"][metadata_type]
                    ).items():
                        host_vars_struct[k] = ValueWithSource(v, source)
            elif isinstance(metadata_value, list):
                for item in metadata_value:
                    logger.debug(f"Processing metadata {metadata_type}/{item}")
                    source = f"{metadata_type}/{item}"
                    if metadata.lookup(metadata_type, item):
                        for k, v in metadata.lookup(metadata_type, item).items():
                            host_vars_struct[k] = ValueWithSource(v, source)
            else:
                raise ValueError(
                    f"Invalid metadata type {type(metadata_value)} ({metadata_type}/{metadata_value})"
                )

        source = f"hosts/{host.stem}"
        for k, v in host_vars.items():
            host_vars_struct[k] = ValueWithSource(v, source)

        previous_source: str | None = None
        for k, v in host_vars_struct.items():
            if previous_source is None:
                f.write(f"# {v.source}\n")
                previous_source = v.source
            elif previous_source != v.source:
                f.write(f"\n# {v.source}\n")
                previous_source = v.source

            save_yaml(f, {k: v.value})

        # Test if generated file is valid yaml
        try:
            load_yaml_cached(f)
        except yaml.YAMLError as e:
            raise ValueError(f"Error reading generated file for host {host.stem}: {e}")

        f.seek(0)
        return f.read()


def get_generated_host_path(base_dir: Path, host: str) -> Path:
    return base_dir.joinpath(f"generated/{host}.yaml")


def get_all_host_files(base_path: Path) -> list[Path]:
    host_files = base_path.joinpath("hosts/").rglob("*.yaml")
    return [f for f in host_files if f.is_file()]


@dataclass
class GeneratedHost:
    """Represents a generated host file"""

    name: str
    vars: dict


def get_all_generated_hosts(base_path: Path) -> list[GeneratedHost]:
    host_files = base_path.joinpath("generated/").rglob("*.yaml")
    return [
        GeneratedHost(name=f.stem, vars=load_yaml(f)) for f in host_files if f.is_file()
    ]



from pathlib import Path
from invgen.files import load_yaml
from invgen.logging import logger


class MetadataVars:
    def __init__(self):
        self._metadata = {}

    def add_metadata(self, name: str):
        logger.info(f"Adding metadata type {name}")
        self._metadata[name] = {}

    def set_vars(self, metadata_type: str, name: str, vars: dict):
        self._metadata[metadata_type][name] = vars

    def __repr__(self) -> str:
        return f"MetadataVars({self._metadata})"

    def __getattr__(self, name: str):
        if name in self._metadata:
            return self._metadata[name]
        raise AttributeError(f"'MetadataVars' object has no attribute '{name}'")

    def lookup(self, metadata: str, key: str) -> dict:
        """
        raises ValueError if metadata or key not found
        """

        logger.debug(f"Looking up metadata {metadata}/{key}")
        try:
            sub_metadata = dict(getattr(self, metadata))
        except AttributeError:
            raise ValueError(
                f'Metadata type "{metadata}" not found. Consider adding a directory at "metadata/{metadata}"'
            )

        if key not in sub_metadata:
            logger.warning(
                f"Metadata {metadata}/{key} not found. "
                f'Did you forget to add "{key}.yaml" to "metadata/{metadata}/"?'
            )

        return sub_metadata.get(key, {})


def build_metadata_vars(data_dir: Path) -> MetadataVars:
    vars = MetadataVars()
    logger.info("Getting metadata vars")
    metadata_dir = data_dir.joinpath("metadata")
    for subdir in metadata_dir.iterdir():
        if subdir.is_dir():
            vars.add_metadata(subdir.name)
            for file in subdir.rglob("*.yaml"):
                vars.set_vars(subdir.name, file.stem, load_yaml(file))
    return vars



from io import TextIOWrapper
from invgen.logging import logger
from pathlib import Path
import yaml
from functools import lru_cache

try:
    from yaml import CSafeLoader as SafeLoader
    from yaml import CSafeDumper as SafeDumper
except ImportError as e:
    logger.error(e)
    from yaml import SafeLoader, SafeDumper


class VaultPass(str):
    """String subclass to represent vault-encrypted values"""

    pass


def ansible_vault_constructor(loader, node: yaml.nodes.ScalarNode) -> VaultPass:
    """Construct a VaultPass object from a YAML node"""
    if not isinstance(node, yaml.ScalarNode):
        raise yaml.constructor.ConstructorError(
            None, None, f"expected a scalar node, but found {node.id}", node.start_mark
        )
    return VaultPass(loader.construct_scalar(node))


def ansible_vault_representer(dumper, data: VaultPass) -> yaml.nodes.ScalarNode:
    """Represent a VaultPass object as a YAML scalar node"""
    return dumper.represent_scalar(
        "!vault",
        str(data),
        style="|",  # Use literal block style for multiline vault strings
    )


# Register the custom constructor and representer
SafeLoader.add_constructor("!vault", ansible_vault_constructor)
SafeDumper.add_representer(VaultPass, ansible_vault_representer)


@lru_cache
def load_yaml_cached(file: Path) -> dict:
    """Load a yaml file and cache the result"""
    return load_yaml(file)


def load_yaml(file: Path | TextIOWrapper) -> dict:
    """Load a yaml file"""
    try:
        if isinstance(file, Path):
            return yaml.load(file.read_text(), Loader=SafeLoader)
        else:
            return yaml.load(file.read(), Loader=SafeLoader)
    except (yaml.YAMLError, OSError) as e:
        logger.error(f"Error processing file: {e}")
        raise


def save_yaml(file: Path | TextIOWrapper, data: dict, sort_keys: bool = False) -> None:
    """Save a yaml file"""
    try:
        content = yaml.dump(
            data,
            Dumper=SafeDumper,
            sort_keys=sort_keys,
            indent=2,
            default_flow_style=False,
            allow_unicode=True,
        )
        if isinstance(file, Path):
            file.write_text(content)
        else:
            file.write(content)
    except (yaml.YAMLError, OSError) as e:
        logger.error(f"Error processing file: {e}")
        raise



from jinja2 import Environment, FileSystemLoader
from pathlib import Path

import jinja2


def render_template(template_path: Path, undefined_error: bool = True, **kwargs) -> str:
    env = Environment(loader=FileSystemLoader(template_path.parent))
    if undefined_error:
        env.undefined = jinja2.StrictUndefined

    template = env.get_template(template_path.name)
    return template.render(**kwargs)
